@@ -1,5 +1,5 @@
 package Positron::Environment;
-our $VERSION = 'v0.0.5'; # VERSION
+our $VERSION = 'v0.0.6'; # VERSION
 
 =head1 NAME
 
@@ -7,7 +7,7 @@ Positron::Environment - container class for template parameters
 
 =head1 VERSION
 
-version v0.0.5
+version v0.0.6
 
 =head1 SYNOPSIS
 
@@ -126,6 +126,7 @@ sub get {
         return $self->{'data'}->{$key};
     } elsif (ref($self->{'data'}) eq 'ARRAY') {
         # What about parents with array refs?
+        no warnings 'numeric'; # all else is 0, that's ok.
         return $self->{'data'}->[int($key)];
     # N.B.: other scalars (non-refs, objects) never perform subqueries, always 'undef'
     } elsif ($self->{'parent'}) {
@@ -157,6 +158,7 @@ sub set {
     if ($key eq '_') {
         $self->{'data'} = $value;
     } elsif (ref($self->{'data'}) eq 'ARRAY') {
+        no warnings 'numeric';
         $self->{'data'}->[int($key)] = $value;
     } elsif (ref($self->{'data'}) eq 'HASH') {
         $self->{'data'}->{$key} = $value;
